"""Build OSC bundles for client applications."""

from typing import List

from pythonosc import osc_bundle
from pythonosc import osc_message
from pythonosc.parsing import osc_types

# Shortcut to specify an immediate execution of messages in the bundle.
IMMEDIATELY = osc_types.IMMEDIATELY


class BuildError(Exception):
    """Error raised when an error occurs building the bundle."""


class OscBundleBuilder(object):
    """Builds arbitrary OscBundle instances."""

    def __init__(self, timestamp: int) -> None:
        """Build a new bundle with the associated timestamp.

        Args:
          - timestamp: system time represented as a floating point number of
                       seconds since the epoch in UTC or IMMEDIATELY.
        """
        self._timestamp = timestamp
        self._contents = []  # type: List[osc_bundle.OscBundle]

    def add_content(self, content: osc_bundle.OscBundle) -> None:
        """Add a new content to this bundle.

        Args:
          - content: Either an OscBundle or an OscMessage
        """
        self._contents.append(content)

    def build(self) -> osc_bundle.OscBundle:
        """Build an OscBundle with the current state of this builder.

        Raises:
          - BuildError: if we could not build the bundle.
        """
        dgram = b'#bundle\x00'
        try:
            dgram += osc_types.write_date(self._timestamp)
            for content in self._contents:
                if (type(content) == osc_message.OscMessage
                        or type(content) == osc_bundle.OscBundle):
                    size = content.size
                    dgram += osc_types.write_int(size)
                    dgram += content.dgram
                else:
                    raise BuildError(
                        "Content must be either OscBundle or OscMessage"
                        "found {}".format(type(content)))
            return osc_bundle.OscBundle(dgram)
        except osc_types.BuildError as be:
            raise BuildError('Could not build the bundle {}'.format(be))



"""OSC Servers that receive UDP packets and invoke handlers accordingly.
"""

import asyncio
import os
import socketserver

from pythonosc import osc_bundle
from pythonosc import osc_message
from pythonosc.dispatcher import Dispatcher

from asyncio import BaseEventLoop

from socket import socket as _socket
from typing import Any, Tuple, Union, cast, Coroutine

_RequestType = Union[_socket, Tuple[bytes, _socket]]
_AddressType = Union[Tuple[str, int], str]


class _UDPHandler(socketserver.BaseRequestHandler):
    """Handles correct UDP messages for all types of server."""

    def handle(self) -> None:
        """Calls the handlers via dispatcher

        This method is called after a basic sanity check was done on the datagram,
        whether this datagram looks like an osc message or bundle.
        If not the server won't call it and so no new
        threads/processes will be spawned.
        """
        server = cast(OSCUDPServer, self.server)
        server.dispatcher.call_handlers_for_packet(self.request[0], self.client_address)


def _is_valid_request(request: _RequestType) -> bool:
    """Returns true if the request's data looks like an osc bundle or message.

    Returns:
        True if request is OSC bundle or OSC message
    """
    assert isinstance(request, tuple)  # TODO: handle requests which are passed just as a socket?
    data = request[0]
    return (
            osc_bundle.OscBundle.dgram_is_bundle(data)
            or osc_message.OscMessage.dgram_is_message(data))


class OSCUDPServer(socketserver.UDPServer):
    """Superclass for different flavors of OSC UDP servers"""

    def __init__(self, server_address: Tuple[str, int], dispatcher: Dispatcher, bind_and_activate: bool = True) -> None:
        """Initialize

        Args:
            server_address: IP and port of server
            dispatcher: Dispatcher this server will use
            (optional) bind_and_activate: default=True defines if the server has to start on call of constructor  
        """
        super().__init__(server_address, _UDPHandler, bind_and_activate)
        self._dispatcher = dispatcher

    def verify_request(self, request: _RequestType, client_address: _AddressType) -> bool:
        """Returns true if the data looks like a valid OSC UDP datagram

        Args:
            request: Incoming data
            client_address: IP and port of client this message came from

        Returns:
            True if request is OSC bundle or OSC message
        """
        return _is_valid_request(request)

    @property
    def dispatcher(self) -> Dispatcher:
        return self._dispatcher


class BlockingOSCUDPServer(OSCUDPServer):
    """Blocking version of the UDP server.

    Each message will be handled sequentially on the same thread.
    Use this is you don't care about latency in your message handling or don't
    have a multiprocess/multithread environment.
    """


class ThreadingOSCUDPServer(socketserver.ThreadingMixIn, OSCUDPServer):
    """Threading version of the OSC UDP server.

    Each message will be handled in its own new thread.
    Use this when lightweight operations are done by each message handlers.
    """


if hasattr(os, "fork"):
    class ForkingOSCUDPServer(socketserver.ForkingMixIn, OSCUDPServer):
        """Forking version of the OSC UDP server.

        Each message will be handled in its own new process.
        Use this when heavyweight operations are done by each message handlers
        and forking a whole new process for each of them is worth it.
        """


class AsyncIOOSCUDPServer():
    """Asynchronous OSC Server

    An asynchronous OSC Server using UDP. It creates a datagram endpoint that runs in an event loop.
    """

    def __init__(self, server_address: Tuple[str, int], dispatcher: Dispatcher, loop: BaseEventLoop) -> None:
        """Initialize

        Args:
            server_address: IP and port of server
            dispatcher: Dispatcher this server shall use
            loop: Event loop to add the server task to. Use ``asyncio.get_event_loop()`` unless you know what you're
                doing.
        """

        self._server_address = server_address
        self._dispatcher = dispatcher
        self._loop = loop

    class _OSCProtocolFactory(asyncio.DatagramProtocol):
        """OSC protocol factory which passes datagrams to dispatcher"""

        def __init__(self, dispatcher: Dispatcher) -> None:
            self.dispatcher = dispatcher

        def datagram_received(self, data: bytes, client_address: Tuple[str, int]) -> None:
            self.dispatcher.call_handlers_for_packet(data, client_address)

    def serve(self) -> None:
        """Creates a datagram endpoint and registers it with event loop.

        Use this only in synchronous code (i.e. not from within a coroutine). This will start the server and run it
        forever or until a ``stop()`` is called on the event loop.
        """
        self._loop.run_until_complete(self.create_serve_endpoint())

    def create_serve_endpoint(self) -> Coroutine[Any, Any, Tuple[asyncio.transports.BaseTransport, asyncio.DatagramProtocol]]:
        """Creates a datagram endpoint and registers it with event loop as coroutine.

        Returns:
            Awaitable coroutine that returns transport and protocol objects
        """
        return self._loop.create_datagram_endpoint(
            lambda: self._OSCProtocolFactory(self.dispatcher),
            local_addr=self._server_address)

    @property
    def dispatcher(self) -> Dispatcher:
        return self._dispatcher



"""Unit tests for the osc_types module."""
import unittest

from pythonosc.parsing import ntp
from pythonosc.parsing import osc_types

from datetime import datetime


class TestString(unittest.TestCase):
    def test_get_string(self):
        cases = {
            b"A\x00\x00\x00": ("A", 4),
            b"AB\x00\x00": ("AB", 4),
            b"ABC\x00": ("ABC", 4),
            b"ABCD\x00\x00\x00\x00": ("ABCD", 8),

            b"ABCD\x00\x00\x00\x00GARBAGE": ("ABCD", 8),
            b'\x00\x00\x00\x00': ("", 4),
        }

        for dgram, expected in cases.items():
            self.assertEqual(expected, osc_types.get_string(dgram, 0))

    def test_get_string_raises_on_wrong_dgram(self):
        cases = [
            b'blablaba',
            b'',
            b'\x00',
            b'\x00\x00',
            True,
        ]

        for case in cases:
            self.assertRaises(
                osc_types.ParseError, osc_types.get_string, case, 0)

    def test_get_string_raises_when_datagram_too_short(self):
        self.assertRaises(
            osc_types.ParseError, osc_types.get_string, b'abc\x00', 1)

    def test_get_string_raises_on_wrong_start_index_negative(self):
        self.assertRaises(
            osc_types.ParseError, osc_types.get_string, b'abc\x00', -1)


class TestInteger(unittest.TestCase):
    def test_get_integer(self):
        cases = {
            b"\x00\x00\x00\x00": (0, 4),
            b"\x00\x00\x00\x01": (1, 4),
            b"\x00\x00\x00\x02": (2, 4),
            b"\x00\x00\x00\x03": (3, 4),

            b"\x00\x00\x01\x00": (256, 4),
            b"\x00\x01\x00\x00": (65536, 4),
            b"\x01\x00\x00\x00": (16777216, 4),

            b"\x00\x00\x00\x01GARBAGE": (1, 4),
        }

        for dgram, expected in cases.items():
            self.assertEqual(
                expected, osc_types.get_int(dgram, 0))

    def test_get_integer_raises_on_type_error(self):
        cases = [b'', True]

        for case in cases:
            self.assertRaises(osc_types.ParseError, osc_types.get_int, case, 0)

    def test_get_integer_raises_on_wrong_start_index(self):
        self.assertRaises(
            osc_types.ParseError, osc_types.get_int, b'\x00\x00\x00\x11', 1)

    def test_get_integer_raises_on_wrong_start_index_negative(self):
        self.assertRaises(
            osc_types.ParseError, osc_types.get_int, b'\x00\x00\x00\x00', -1)

    def test_datagram_too_short(self):
        dgram = b'\x00' * 3
        self.assertRaises(osc_types.ParseError, osc_types.get_int, dgram, 2)


class TestRGBA(unittest.TestCase):
    def test_get_rgba(self):
        cases = {
            b"\x00\x00\x00\x00": (0, 4),
            b"\x00\x00\x00\x01": (1, 4),
            b"\x00\x00\x00\x02": (2, 4),
            b"\x00\x00\x00\x03": (3, 4),

            b"\xFF\x00\x00\x00": (4278190080, 4),
            b"\x00\xFF\x00\x00": (16711680, 4),
            b"\x00\x00\xFF\x00": (65280, 4),
            b"\x00\x00\x00\xFF": (255, 4),

            b"\x00\x00\x00\x01GARBAGE": (1, 4),
        }

        for dgram, expected in cases.items():
            self.assertEqual(
                expected, osc_types.get_rgba(dgram, 0))

    def test_get_rgba_raises_on_type_error(self):
        cases = [b'', True]

        for case in cases:
            self.assertRaises(osc_types.ParseError, osc_types.get_rgba, case, 0)

    def test_get_rgba_raises_on_wrong_start_index(self):
        self.assertRaises(
            osc_types.ParseError, osc_types.get_rgba, b'\x00\x00\x00\x11', 1)

    def test_get_rgba_raises_on_wrong_start_index_negative(self):
        self.assertRaises(
            osc_types.ParseError, osc_types.get_rgba, b'\x00\x00\x00\x00', -1)

    def test_datagram_too_short(self):
        dgram = b'\x00' * 3
        self.assertRaises(osc_types.ParseError, osc_types.get_rgba, dgram, 2)


class TestMidi(unittest.TestCase):
    def test_get_midi(self):
        cases = {
            b"\x00\x00\x00\x00": ((0, 0, 0, 0), 4),
            b"\x00\x00\x00\x02": ((0, 0, 0, 1), 4),
            b"\x00\x00\x00\x02": ((0, 0, 0, 2), 4),
            b"\x00\x00\x00\x03": ((0, 0, 0, 3), 4),

            b"\x00\x00\x01\x00": ((0, 0, 1, 0), 4),
            b"\x00\x01\x00\x00": ((0, 1, 0, 0), 4),
            b"\x01\x00\x00\x00": ((1, 0, 0, 0), 4),

            b"\x00\x00\x00\x01GARBAGE": ((0, 0, 0, 1), 4),
        }

        for dgram, expected in cases.items():
            self.assertEqual(
                expected, osc_types.get_midi(dgram, 0))

    def test_get_midi_raises_on_type_error(self):
        cases = [b'', True]

        for case in cases:
            self.assertRaises(osc_types.ParseError, osc_types.get_midi, case, 0)

    def test_get_midi_raises_on_wrong_start_index(self):
        self.assertRaises(
            osc_types.ParseError, osc_types.get_midi, b'\x00\x00\x00\x11', 1)

    def test_get_midi_raises_on_wrong_start_index_negative(self):
        self.assertRaises(
            osc_types.ParseError, osc_types.get_midi, b'\x00\x00\x00\x00', -1)

    def test_datagram_too_short(self):
        dgram = b'\x00' * 3
        self.assertRaises(osc_types.ParseError, osc_types.get_midi, dgram, 2)


class TestDate(unittest.TestCase):
    def test_get_timetag(self):
        cases = {
            b"\xde\x9c\x91\xbf\x00\x01\x00\x00": ((datetime(2018, 5, 8, 21, 14, 39), 65536), 8),  # NOTE: fraction is expresed as 32bit OSC.
            b"\x00\x00\x00\x00\x00\x00\x00\x00": ((datetime(1900, 1, 1, 0, 0, 0), 0), 8),
            b"\x83\xaa\x7E\x80\x0A\x00\xB0\x0C": ((datetime(1970, 1, 1, 0, 0, 0), 167817228), 8)
        }

        for dgram, expected in cases.items():
            self.assertEqual(expected, osc_types.get_timetag(dgram, 0))

    def test_get_timetag_raises_on_wrong_start_index_negative(self):
        self.assertRaises(
            osc_types.ParseError, osc_types.get_timetag, b'\x00\x00\x00\x00\x00\x00\x00\x00', -1)

    def test_get_timetag_raises_on_type_error(self):
        cases = [b'', True]

        for case in cases:
            self.assertRaises(osc_types.ParseError, osc_types.get_timetag, case, 0)

    def test_get_timetag_raises_on_wrong_start_index(self):
        self.assertRaises(
            osc_types.ParseError, osc_types.get_date, b'\x00\x00\x00\x11\x00\x00\x00\x11', 1)

    def test_ttag_datagram_too_short(self):
        dgram = b'\x00' * 7
        self.assertRaises(osc_types.ParseError, osc_types.get_timetag, dgram, 6)

        dgram = b'\x00' * 2
        self.assertRaises(osc_types.ParseError, osc_types.get_timetag, dgram, 1)

        dgram = b'\x00' * 5
        self.assertRaises(osc_types.ParseError, osc_types.get_timetag, dgram, 4)

        dgram = b'\x00' * 1
        self.assertRaises(osc_types.ParseError, osc_types.get_timetag, dgram, 0)


class TestFloat(unittest.TestCase):
    def test_get_float(self):
        cases = {
            b"\x00\x00\x00\x00": (0.0, 4),
            b"?\x80\x00\x00'": (1.0, 4),
            b'@\x00\x00\x00': (2.0, 4),

            b"\x00\x00\x00\x00GARBAGE": (0.0, 4),
        }

        for dgram, expected in cases.items():
            self.assertAlmostEqual(expected, osc_types.get_float(dgram, 0))

    def test_get_float_raises_on_wrong_dgram(self):
        cases = [True]

        for case in cases:
            self.assertRaises(osc_types.ParseError, osc_types.get_float, case, 0)

    def test_get_float_raises_on_type_error(self):
        cases = [None]

        for case in cases:
            self.assertRaises(osc_types.ParseError, osc_types.get_float, case, 0)

    def test_datagram_too_short_pads(self):
        dgram = b'\x00' * 2
        self.assertEqual((0, 4), osc_types.get_float(dgram, 0))


class TestDouble(unittest.TestCase):
    def test_get_double(self):
        cases = {
            b'\x00\x00\x00\x00\x00\x00\x00\x00': (0.0, 8),
            b'?\xf0\x00\x00\x00\x00\x00\x00': (1.0, 8),
            b'@\x00\x00\x00\x00\x00\x00\x00': (2.0, 8),
            b'\xbf\xf0\x00\x00\x00\x00\x00\x00': (-1.0, 8),
            b'\xc0\x00\x00\x00\x00\x00\x00\x00': (-2.0, 8),

            b"\x00\x00\x00\x00\x00\x00\x00\x00GARBAGE": (0.0, 8),
        }

        for dgram, expected in cases.items():
            self.assertAlmostEqual(expected, osc_types.get_double(dgram, 0))

    def test_get_double_raises_on_wrong_dgram(self):
        cases = [True]

        for case in cases:
            self.assertRaises(osc_types.ParseError, osc_types.get_double, case, 0)

    def test_get_double_raises_on_type_error(self):
        cases = [None]

        for case in cases:
            self.assertRaises(osc_types.ParseError, osc_types.get_double, case, 0)

    def test_datagram_too_short_pads(self):
        dgram = b'\x00' * 2
        self.assertRaises(osc_types.ParseError, osc_types.get_double, dgram, 0)


class TestBlob(unittest.TestCase):
    def test_get_blob(self):
        cases = {
            b"\x00\x00\x00\x00": (b"", 4),
            b"\x00\x00\x00\x08stuff\x00\x00\x00": (b"stuff\x00\x00\x00", 12),
            b"\x00\x00\x00\x04\x00\x00\x00\x00": (b"\x00\x00\x00\x00", 8),
            b"\x00\x00\x00\x02\x00\x00\x00\x00": (b"\x00\x00", 8),

            b"\x00\x00\x00\x08stuff\x00\x00\x00datagramcontinues": (
                b"stuff\x00\x00\x00", 12),
        }

        for dgram, expected in cases.items():
            self.assertEqual(expected, osc_types.get_blob(dgram, 0))

    def test_get_blob_raises_on_wrong_dgram(self):
        cases = [b'', True, b"\x00\x00\x00\x08"]

        for case in cases:
            self.assertRaises(osc_types.ParseError, osc_types.get_blob, case, 0)

    def test_get_blob_raises_on_wrong_start_index(self):
        self.assertRaises(
            osc_types.ParseError, osc_types.get_blob, b'\x00\x00\x00\x11', 1)

    def test_get_blob_raises_too_short_buffer(self):
        self.assertRaises(
            osc_types.ParseError,
            osc_types.get_blob,
            b'\x00\x00\x00\x11\x00\x00', 1)

    def test_get_blog_raises_on_wrong_start_index_negative(self):
        self.assertRaises(
            osc_types.ParseError, osc_types.get_blob, b'\x00\x00\x00\x00', -1)


class TestNTPTimestamp(unittest.TestCase):
    def test_immediately_dgram(self):
        dgram = ntp.IMMEDIATELY
        self.assertEqual(osc_types.IMMEDIATELY, osc_types.get_date(dgram, 0)[0])

    def test_origin_of_time(self):
        dgram = b'\x00' * 8
        self.assertGreater(0, osc_types.get_date(dgram, 0)[0])

    def test_datagram_too_short(self):
        dgram = b'\x00' * 8
        self.assertRaises(osc_types.ParseError, osc_types.get_date, dgram, 2)

    def test_write_date(self):
        time = 1569899476.167749  # known round(time.time(), 6)
        timetag = b'\xe1=BT*\xf1\x98\x00'
        self.assertEqual(timetag, osc_types.write_date(time))


class TestBuildMethods(unittest.TestCase):
    def test_string(self):
        self.assertEqual(b'\x00\x00\x00\x00', osc_types.write_string(''))
        self.assertEqual(b'A\x00\x00\x00', osc_types.write_string('A'))
        self.assertEqual(b'AB\x00\x00', osc_types.write_string('AB'))
        self.assertEqual(b'ABC\x00', osc_types.write_string('ABC'))
        self.assertEqual(b'ABCD\x00\x00\x00\x00', osc_types.write_string('ABCD'))

    def test_string_raises(self):
        self.assertRaises(osc_types.BuildError, osc_types.write_string, 123)

    def test_int(self):
        self.assertEqual(b'\x00\x00\x00\x00', osc_types.write_int(0))
        self.assertEqual(b'\x00\x00\x00\x01', osc_types.write_int(1))

    def test_int_raises(self):
        self.assertRaises(osc_types.BuildError, osc_types.write_int, 'no int')

    def test_float(self):
        self.assertEqual(b'\x00\x00\x00\x00', osc_types.write_float(0.0))
        self.assertEqual(b'?\x00\x00\x00', osc_types.write_float(0.5))
        self.assertEqual(b'?\x80\x00\x00', osc_types.write_float(1.0))
        self.assertEqual(b'?\x80\x00\x00', osc_types.write_float(1))

    def test_float_raises(self):
        self.assertRaises(osc_types.BuildError, osc_types.write_float, 'no float')

    def test_blob(self):
        self.assertEqual(
            b'\x00\x00\x00\x02\x00\x01\x00\x00',
            osc_types.write_blob(b'\x00\x01'))
        self.assertEqual(
            b'\x00\x00\x00\x04\x00\x01\x02\x03',
            osc_types.write_blob(b'\x00\x01\x02\x03'))

    def test_blob_raises(self):
        self.assertRaises(osc_types.BuildError, osc_types.write_blob, b'')


if __name__ == "__main__":
    unittest.main()



import unittest

from pythonosc import osc_packet

_DGRAM_TWO_MESSAGES_IN_BUNDLE = (
    b"#bundle\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x01"
    # First message.
    b"\x00\x00\x00\x10"
    b"/SYNC\x00\x00\x00"
    b",f\x00\x00"
    b"?\x00\x00\x00"
    # Second message, same.
    b"\x00\x00\x00\x10"
    b"/SYNC\x00\x00\x00"
    b",f\x00\x00"
    b"?\x00\x00\x00")

_DGRAM_EMPTY_BUNDLE = (
    b"#bundle\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x01")

_DGRAM_NESTED_MESS = (
    b"#bundle\x00"
    b"\x10\x00\x00\x00\x00\x00\x00\x00"
    # First message.
    b"\x00\x00\x00\x10"  # 16 bytes
    b"/1111\x00\x00\x00"
    b",f\x00\x00"
    b"?\x00\x00\x00"
    # Second message, same.
    b"\x00\x00\x00\x10"  # 16 bytes
    b"/2222\x00\x00\x00"
    b",f\x00\x00"
    b"?\x00\x00\x00"
    # Now another bundle within it, oh my...
    b"\x00\x00\x00$"  # 36 bytes.
    b"#bundle\x00"
    b"\x20\x00\x00\x00\x00\x00\x00\x00"
    # First message.
    b"\x00\x00\x00\x10"
    b"/3333\x00\x00\x00"
    b",f\x00\x00"
    b"?\x00\x00\x00"
    # And another final bundle.
    b"\x00\x00\x00$"  # 36 bytes.
    b"#bundle\x00"
    b"\x15\x00\x00\x00\x00\x00\x00\x01"  # Immediately this one.
    # First message.
    b"\x00\x00\x00\x10"
    b"/4444\x00\x00\x00"
    b",f\x00\x00"
    b"?\x00\x00\x00")


class TestOscPacket(unittest.TestCase):
    def test_two_messages_in_a_bundle(self):
        packet = osc_packet.OscPacket(_DGRAM_TWO_MESSAGES_IN_BUNDLE)
        self.assertEqual(2, len(packet.messages))

    def test_empty_dgram_raises_exception(self):
        self.assertRaises(osc_packet.ParseError, osc_packet.OscPacket, b'')

    def test_empty_bundle(self):
        packet = osc_packet.OscPacket(_DGRAM_EMPTY_BUNDLE)
        self.assertEqual(0, len(packet.messages))

    def test_nested_mess_bundle(self):
        packet = osc_packet.OscPacket(_DGRAM_NESTED_MESS)
        self.assertEqual(4, len(packet.messages))
        self.assertTrue(packet.messages[0][0], packet.messages[1][0])
        self.assertTrue(packet.messages[1][0], packet.messages[2][0])
        self.assertTrue(packet.messages[2][0], packet.messages[3][0])


if __name__ == "__main__":
    unittest.main()



import unittest

from pythonosc import osc_message_builder


class TestOscMessageBuilder(unittest.TestCase):
    def test_just_address(self):
        msg = osc_message_builder.OscMessageBuilder("/a/b/c").build()
        self.assertEqual("/a/b/c", msg.address)
        self.assertEqual([], msg.params)
        # Messages with just an address should still contain the ",".
        self.assertEqual(b'/a/b/c\x00\x00,\x00\x00\x00', msg.dgram)

    def test_no_address_raises(self):
        builder = osc_message_builder.OscMessageBuilder("")
        self.assertRaises(osc_message_builder.BuildError, builder.build)

    def test_wrong_param_raise(self):
        builder = osc_message_builder.OscMessageBuilder("")
        self.assertRaises(ValueError, builder.add_arg, "what?", 1)

    def test_add_arg_invalid_infered_type(self):
        builder = osc_message_builder.OscMessageBuilder('')
        self.assertRaises(ValueError, builder.add_arg, {'name': 'John'})

    def test_all_param_types(self):
        builder = osc_message_builder.OscMessageBuilder(address="/SYNC")
        builder.add_arg(4.0)
        builder.add_arg(2)
        builder.add_arg(1099511627776)
        builder.add_arg("value")
        builder.add_arg(True)
        builder.add_arg(False)
        builder.add_arg(b"\x01\x02\x03")
        builder.add_arg([1, ["abc"]])
        builder.add_arg(None)
        # The same args but with explicit types.
        builder.add_arg(4.0, builder.ARG_TYPE_FLOAT)
        builder.add_arg(2, builder.ARG_TYPE_INT)
        builder.add_arg(1099511627776, builder.ARG_TYPE_INT64)
        builder.add_arg("value", builder.ARG_TYPE_STRING)
        builder.add_arg(True)
        builder.add_arg(False)
        builder.add_arg(b"\x01\x02\x03", builder.ARG_TYPE_BLOB)
        builder.add_arg([1, ["abc"]], [builder.ARG_TYPE_INT, [builder.ARG_TYPE_STRING]])
        builder.add_arg(None, builder.ARG_TYPE_NIL)
        builder.add_arg(4278255360, builder.ARG_TYPE_RGBA)
        builder.add_arg((1, 145, 36, 125), builder.ARG_TYPE_MIDI)
        builder.add_arg(1e-9, builder.ARG_TYPE_DOUBLE)
        self.assertEqual(len("fihsTFb[i[s]]N") * 2 + 3, len(builder.args))
        self.assertEqual("/SYNC", builder.address)
        builder.address = '/SEEK'
        msg = builder.build()
        self.assertEqual("/SEEK", msg.address)
        self.assertSequenceEqual(
            [4.0, 2, 1099511627776, "value", True, False, b"\x01\x02\x03", [1, ["abc"]], None] * 2 +
            [4278255360, (1, 145, 36, 125), 1e-9],
            msg.params)

    def test_long_list(self):
        huge_list = list(range(512))
        builder = osc_message_builder.OscMessageBuilder(address="/SYNC")
        builder.add_arg(huge_list)
        msg = builder.build()
        print(msg._dgram)
        self.assertSequenceEqual([huge_list], msg.params)

    def test_build_wrong_type_raises(self):
        builder = osc_message_builder.OscMessageBuilder(address="/SYNC")
        builder.add_arg('this is not a float', builder.ARG_TYPE_FLOAT)
        self.assertRaises(osc_message_builder.BuildError, builder.build)

    def test_build_noarg_message(self):
        msg = osc_message_builder.OscMessageBuilder(address='/SYNC').build()
        # This reference message was generated with Cycling 74's Max software
        # and then was intercepted with Wireshark
        reference = bytearray.fromhex('2f53594e430000002c000000')
        self.assertSequenceEqual(msg._dgram, reference)

    def test_bool_encoding(self):
        builder = osc_message_builder.OscMessageBuilder('')
        builder.add_arg(0)
        builder.add_arg(1)
        builder.add_arg(False)
        builder.add_arg(True)
        self.assertEqual(builder.args, [("i", 0), ("i", 1), ("F", False), ("T", True)])


if __name__ == "__main__":
    unittest.main()



"""Build OSC messages for client applications."""

from pythonosc import osc_message
from pythonosc.parsing import osc_types

from typing import List, Tuple, Union, Any, Optional

ArgValue = Union[str, bytes, bool, int, float, osc_types.MidiPacket, list]

class BuildError(Exception):
    """Error raised when an incomplete message is trying to be built."""

class OscMessageBuilder(object):
    """Builds arbitrary OscMessage instances."""

    ARG_TYPE_FLOAT = "f"
    ARG_TYPE_DOUBLE = "d"
    ARG_TYPE_INT = "i"
    ARG_TYPE_INT64 = "h"
    ARG_TYPE_STRING = "s"
    ARG_TYPE_BLOB = "b"
    ARG_TYPE_RGBA = "r"
    ARG_TYPE_MIDI = "m"
    ARG_TYPE_TRUE = "T"
    ARG_TYPE_FALSE = "F"
    ARG_TYPE_NIL = "N"

    ARG_TYPE_ARRAY_START = "["
    ARG_TYPE_ARRAY_STOP = "]"

    _SUPPORTED_ARG_TYPES = (
        ARG_TYPE_FLOAT, ARG_TYPE_DOUBLE, ARG_TYPE_INT, ARG_TYPE_INT64, ARG_TYPE_BLOB, ARG_TYPE_STRING,
        ARG_TYPE_RGBA, ARG_TYPE_MIDI, ARG_TYPE_TRUE, ARG_TYPE_FALSE, ARG_TYPE_NIL)

    def __init__(self, address: Optional[str] = None) -> None:
        """Initialize a new builder for a message.

        Args:
          - address: The osc address to send this message to.
        """
        self._address = address
        self._args = []  # type: List[Tuple[str, Union[ArgValue, None]]]

    @property
    def address(self) -> Optional[str]:
        """Returns the OSC address this message will be sent to."""
        return self._address

    @address.setter
    def address(self, value: str) -> None:
        """Sets the OSC address this message will be sent to."""
        self._address = value

    @property
    def args(self) -> List[Tuple[str, Union[ArgValue, None]]]:
        """Returns the (type, value) arguments list of this message."""
        return self._args

    def _valid_type(self, arg_type: str) -> bool:
        if arg_type in self._SUPPORTED_ARG_TYPES:
            return True
        elif isinstance(arg_type, list):
            for sub_type in arg_type:
                if not self._valid_type(sub_type):
                    return False
            return True
        return False

    def add_arg(self, arg_value: ArgValue, arg_type: Optional[str] = None) -> None:
        """Add a typed argument to this message.

        Args:
          - arg_value: The corresponding value for the argument.
          - arg_type: A value in ARG_TYPE_* defined in this class,
                      if none then the type will be guessed.
        Raises:
          - ValueError: if the type is not supported.
        """
        if arg_type and not self._valid_type(arg_type):
            raise ValueError(
                'arg_type must be one of {}, or an array of valid types'
                    .format(self._SUPPORTED_ARG_TYPES))
        if not arg_type:
            arg_type = self._get_arg_type(arg_value)
        if isinstance(arg_type, list):
            self._args.append((self.ARG_TYPE_ARRAY_START, None))
            for v, t in zip(arg_value, arg_type):  # type: ignore[var-annotated, arg-type]
                self.add_arg(v, t)
            self._args.append((self.ARG_TYPE_ARRAY_STOP, None))
        else:
            self._args.append((arg_type, arg_value))

    # The return type here is actually Union[str, List[<self>]], however there
    # is no annotation for a recursive type like this.
    def _get_arg_type(self, arg_value: ArgValue) -> Union[str, Any]:
        """Guess the type of a value.

        Args:
          - arg_value: The value to guess the type of.
        Raises:
          - ValueError: if the type is not supported.
        """
        if isinstance(arg_value, str):
            arg_type = self.ARG_TYPE_STRING  # type: Union[str, Any]
        elif isinstance(arg_value, bytes):
            arg_type = self.ARG_TYPE_BLOB
        elif arg_value is True:
            arg_type = self.ARG_TYPE_TRUE
        elif arg_value is False:
            arg_type = self.ARG_TYPE_FALSE
        elif isinstance(arg_value, int):
            if arg_value.bit_length() > 32:
                arg_type = self.ARG_TYPE_INT64
            else:
                arg_type = self.ARG_TYPE_INT
        elif isinstance(arg_value, float):
            arg_type = self.ARG_TYPE_FLOAT
        elif isinstance(arg_value, tuple) and len(arg_value) == 4:
            arg_type = self.ARG_TYPE_MIDI
        elif isinstance(arg_value, list):
            arg_type = [self._get_arg_type(v) for v in arg_value]
        elif arg_value is None:
            arg_type = self.ARG_TYPE_NIL
        else:
            raise ValueError('Infered arg_value type is not supported')
        return arg_type

    def build(self) -> osc_message.OscMessage:
        """Builds an OscMessage from the current state of this builder.

        Raises:
          - BuildError: if the message could not be build or if the address
                        was empty.

        Returns:
          - an osc_message.OscMessage instance.
        """
        if not self._address:
            raise BuildError('OSC addresses cannot be empty')
        dgram = b''
        try:
            # Write the address.
            dgram += osc_types.write_string(self._address)
            if not self._args:
                dgram += osc_types.write_string(',')
                return osc_message.OscMessage(dgram)

            # Write the parameters.
            arg_types = "".join([arg[0] for arg in self._args])
            dgram += osc_types.write_string(',' + arg_types)
            for arg_type, value in self._args:
                if arg_type == self.ARG_TYPE_STRING:
                    dgram += osc_types.write_string(value)  # type: ignore[arg-type]
                elif arg_type == self.ARG_TYPE_INT:
                    dgram += osc_types.write_int(value)  # type: ignore[arg-type]
                elif arg_type == self.ARG_TYPE_INT64:
                    dgram += osc_types.write_int64(value)  # type: ignore[arg-type]
                elif arg_type == self.ARG_TYPE_FLOAT:
                    dgram += osc_types.write_float(value)  # type: ignore[arg-type]
                elif arg_type == self.ARG_TYPE_DOUBLE:
                    dgram += osc_types.write_double(value)  # type: ignore[arg-type]
                elif arg_type == self.ARG_TYPE_BLOB:
                    dgram += osc_types.write_blob(value)  # type: ignore[arg-type]
                elif arg_type == self.ARG_TYPE_RGBA:
                    dgram += osc_types.write_rgba(value)  # type: ignore[arg-type]
                elif arg_type == self.ARG_TYPE_MIDI:
                    dgram += osc_types.write_midi(value)  # type: ignore[arg-type]
                elif arg_type in (self.ARG_TYPE_TRUE,
                                  self.ARG_TYPE_FALSE,
                                  self.ARG_TYPE_ARRAY_START,
                                  self.ARG_TYPE_ARRAY_STOP,
                                  self.ARG_TYPE_NIL):
                    continue
                else:
                    raise BuildError('Incorrect parameter type found {}'.format(
                        arg_type))

            return osc_message.OscMessage(dgram)
        except osc_types.BuildError as be:
            raise BuildError('Could not build the message: {}'.format(be))



import unittest

from pythonosc import osc_bundle_builder
from pythonosc import osc_message_builder


class TestOscBundleBuilder(unittest.TestCase):
    def test_empty_bundle(self):
        bundle = osc_bundle_builder.OscBundleBuilder(
            osc_bundle_builder.IMMEDIATELY).build()
        self.assertEqual(0, bundle.num_contents)

    def test_raises_on_build(self):
        bundle = osc_bundle_builder.OscBundleBuilder(0.0)
        bundle.add_content(None)
        self.assertRaises(osc_bundle_builder.BuildError, bundle.build)

    def test_raises_on_invalid_timestamp(self):
        bundle = osc_bundle_builder.OscBundleBuilder("I am not a timestamp")
        self.assertRaises(osc_bundle_builder.BuildError, bundle.build)

    def test_build_complex_bundle(self):
        bundle = osc_bundle_builder.OscBundleBuilder(
            osc_bundle_builder.IMMEDIATELY)
        msg = osc_message_builder.OscMessageBuilder(address="/SYNC")
        msg.add_arg(4.0)
        # Add 4 messages in the bundle, each with more arguments.
        bundle.add_content(msg.build())
        msg.add_arg(2)
        bundle.add_content(msg.build())
        msg.add_arg("value")
        bundle.add_content(msg.build())
        msg.add_arg(b"\x01\x02\x03")
        bundle.add_content(msg.build())

        sub_bundle = bundle.build()
        # Now add the same bundle inside itself.
        bundle.add_content(sub_bundle)

        bundle = bundle.build()
        self.assertEqual(5, bundle.num_contents)


if __name__ == "__main__":
    unittest.main()



import unittest

from pythonosc import osc_message
from pythonosc import osc_bundle
from pythonosc.parsing import osc_types

_DGRAM_KNOB_ROTATES_BUNDLE = (
    b"#bundle\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x01"
    b"\x00\x00\x00\x14"
    b"/LFO_Rate\x00\x00\x00"
    b",f\x00\x00"
    b">\x8c\xcc\xcd")

_DGRAM_SWITCH_GOES_OFF = (
    b"#bundle\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x01"
    b"\x00\x00\x00\x10"
    b"/SYNC\x00\x00\x00"
    b",f\x00\x00"
    b"\x00\x00\x00\x00")

_DGRAM_SWITCH_GOES_ON = (
    b"#bundle\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x01"
    b"\x00\x00\x00\x10"
    b"/SYNC\x00\x00\x00"
    b",f\x00\x00"
    b"?\x00\x00\x00")

_DGRAM_TWO_MESSAGES_IN_BUNDLE = (
    b"#bundle\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x01"
    # First message.
    b"\x00\x00\x00\x10"
    b"/SYNC\x00\x00\x00"
    b",f\x00\x00"
    b"?\x00\x00\x00"
    # Second message, same.
    b"\x00\x00\x00\x10"
    b"/SYNC\x00\x00\x00"
    b",f\x00\x00"
    b"?\x00\x00\x00")

_DGRAM_EMPTY_BUNDLE = (
    b"#bundle\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x01")

_DGRAM_BUNDLE_IN_BUNDLE = (
    b"#bundle\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x01"
    b"\x00\x00\x00("  # length of sub bundle: 40 bytes.
    b"#bundle\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x01"
    b"\x00\x00\x00\x10"
    b"/SYNC\x00\x00\x00"
    b",f\x00\x00"
    b"?\x00\x00\x00")

_DGRAM_INVALID = (
    b"#bundle\x00"
    b"\x00\x00\x00")

_DGRAM_INVALID_INDEX = (
    b"#bundle\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x01"
    b"\x00\x00\x00\x20"
    b"/SYNC\x00\x00\x00\x00")

_DGRAM_UNKNOWN_TYPE = (
    b"#bundle\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x01"
    b"\x00\x00\x00\x10"
    b"iamnotaslash")


class TestOscBundle(unittest.TestCase):
    def test_switch_goes_off(self):
        bundle = osc_bundle.OscBundle(_DGRAM_SWITCH_GOES_OFF)
        self.assertEqual(1, bundle.num_contents)
        self.assertEqual(len(_DGRAM_SWITCH_GOES_OFF), bundle.size)
        self.assertEqual(osc_types.IMMEDIATELY, bundle.timestamp)

    def test_switch_goes_on(self):
        bundle = osc_bundle.OscBundle(_DGRAM_SWITCH_GOES_ON)
        self.assertEqual(1, bundle.num_contents)
        self.assertEqual(len(_DGRAM_SWITCH_GOES_ON), bundle.size)
        self.assertEqual(osc_types.IMMEDIATELY, bundle.timestamp)

    def test_datagram_length(self):
        bundle = osc_bundle.OscBundle(_DGRAM_KNOB_ROTATES_BUNDLE)
        self.assertEqual(1, bundle.num_contents)
        self.assertEqual(len(_DGRAM_KNOB_ROTATES_BUNDLE), bundle.size)
        self.assertEqual(osc_types.IMMEDIATELY, bundle.timestamp)

    def test_two_messages_in_bundle(self):
        bundle = osc_bundle.OscBundle(_DGRAM_TWO_MESSAGES_IN_BUNDLE)
        self.assertEqual(2, bundle.num_contents)
        self.assertEqual(osc_types.IMMEDIATELY, bundle.timestamp)
        for content in bundle:
            self.assertEqual(osc_message.OscMessage, type(content))

    def test_empty_bundle(self):
        bundle = osc_bundle.OscBundle(_DGRAM_EMPTY_BUNDLE)
        self.assertEqual(0, bundle.num_contents)
        self.assertEqual(osc_types.IMMEDIATELY, bundle.timestamp)

    def test_bundle_in_bundle_we_must_go_deeper(self):
        bundle = osc_bundle.OscBundle(_DGRAM_BUNDLE_IN_BUNDLE)
        self.assertEqual(1, bundle.num_contents)
        self.assertEqual(osc_types.IMMEDIATELY, bundle.timestamp)
        self.assertEqual(osc_bundle.OscBundle, type(bundle.content(0)))

    def test_dgram_is_bundle(self):
        self.assertTrue(osc_bundle.OscBundle.dgram_is_bundle(
            _DGRAM_SWITCH_GOES_ON))
        self.assertFalse(osc_bundle.OscBundle.dgram_is_bundle(b'junk'))

    def test_raises_on_invalid_datagram(self):
        self.assertRaises(
            osc_bundle.ParseError, osc_bundle.OscBundle, _DGRAM_INVALID)
        self.assertRaises(
            osc_bundle.ParseError, osc_bundle.OscBundle, _DGRAM_INVALID_INDEX)

    def test_unknown_type(self):
        bundle = osc_bundle.OscBundle(_DGRAM_UNKNOWN_TYPE)


if __name__ == "__main__":
    unittest.main()



"""UDP Clients for sending OSC messages to an OSC server"""

import sys

if sys.version_info > (3, 5):
    from collections.abc import Iterable
else:
    from collections import Iterable

import socket

from .osc_message_builder import OscMessageBuilder, ArgValue
from pythonosc.osc_message import OscMessage
from pythonosc.osc_bundle import OscBundle

from typing import Union

class UDPClient(object):
    """OSC client to send :class:`OscMessage` or :class:`OscBundle` via UDP"""

    def __init__(self, address: str, port: int, allow_broadcast: bool = False, family: socket.AddressFamily = socket.AF_UNSPEC) -> None:
        """Initialize client

        As this is UDP it will not actually make any attempt to connect to the
        given server at ip:port until the send() method is called.

        Args:
            address: IP address of server
            port: Port of server
            allow_broadcast: Allow for broadcast transmissions
            family: address family parameter (passed to socket.getaddrinfo)
        """

        for addr in socket.getaddrinfo(address, port, type=socket.SOCK_DGRAM, family=family):
            af, socktype, protocol, canonname, sa = addr

            try:
                self._sock = socket.socket(af, socktype)
            except OSError:
                continue
            break

        self._sock.setblocking(False)
        if allow_broadcast:
            self._sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
        self._address = address
        self._port = port

    def send(self, content: Union[OscMessage, OscBundle]) -> None:
        """Sends an :class:`OscMessage` or :class:`OscBundle` via UDP

        Args:
            content: Message or bundle to be sent
        """
        self._sock.sendto(content.dgram, (self._address, self._port))


class SimpleUDPClient(UDPClient):
    """Simple OSC client that automatically builds :class:`OscMessage` from arguments"""

    def send_message(self, address: str, value: ArgValue) -> None:
        """Build :class:`OscMessage` from arguments and send to server

        Args:
            address: OSC address the message shall go to
            value: One or more arguments to be added to the message
        """
        builder = OscMessageBuilder(address=address)
        if value is None:
            values = []
        elif not isinstance(value, Iterable) or isinstance(value, (str, bytes)):
            values = [value]
        else:
            values = value
        for val in values:
            builder.add_arg(val)
        msg = builder.build()
        self.send(msg)



"""Parsing and conversion of NTP dates contained in datagrams."""

import datetime
import struct
import time

from typing import NamedTuple

# 63 zero bits followed by a one in the least signifigant bit is a special
# case meaning "immediately."
IMMEDIATELY = struct.pack('>Q', 1)

# timetag * (1 / 2 ** 32) == l32bits + (r32bits / 1 ** 32)
_NTP_TIMESTAMP_TO_SECONDS = 1. / 2. ** 32.
_SECONDS_TO_NTP_TIMESTAMP = 2. ** 32.

# From NTP lib.
_SYSTEM_EPOCH = datetime.date(*time.gmtime(0)[0:3])
_NTP_EPOCH = datetime.date(1900, 1, 1)
# _NTP_DELTA is 2208988800
_NTP_DELTA = (_SYSTEM_EPOCH - _NTP_EPOCH).days * 24 * 3600


Timestamp = NamedTuple('Timestamp', [
    ('seconds', int),
    ('fraction', int),
])


class NtpError(Exception):
  """Base class for ntp module errors."""


def parse_timestamp(timestamp: int) -> Timestamp:
    """Parse NTP timestamp as Timetag.
    """
    seconds = timestamp >> 32
    fraction = timestamp & 0xFFFFFFFF
    return Timestamp(seconds, fraction)


def ntp_to_system_time(timestamp: bytes) -> float:
    """Convert a NTP timestamp to system time in seconds.
    """
    try:
        timestamp = struct.unpack('>Q', timestamp)[0]
    except Exception as e:
        raise NtpError(e)
    return timestamp * _NTP_TIMESTAMP_TO_SECONDS - _NTP_DELTA


def system_time_to_ntp(seconds: float) -> bytes:
    """Convert a system time in seconds to NTP timestamp.
    """
    try:
      seconds = seconds + _NTP_DELTA
    except TypeError as e:
      raise NtpError(e)
    return struct.pack('>Q', int(seconds * _SECONDS_TO_NTP_TIMESTAMP))


def ntp_time_to_system_epoch(seconds: float) -> float:
    """Convert a NTP time in seconds to system time in seconds.
    """
    return seconds - _NTP_DELTA


def system_time_to_ntp_epoch(seconds: float) -> float:
    """Convert a system time in seconds to NTP time in seconds.
    """
    return seconds + _NTP_DELTA



import unittest
import unittest.mock

from pythonosc import dispatcher
from pythonosc import osc_server

_SIMPLE_PARAM_INT_MSG = (
    b"/SYNC\x00\x00\x00"
    b",i\x00\x00"
    b"\x00\x00\x00\x04")

# Regression test for a datagram that should NOT be stripped, ever...
_SIMPLE_PARAM_INT_9 = b'/debug\x00\x00,i\x00\x00\x00\x00\x00\t'

_SIMPLE_MSG_NO_PARAMS = b"/SYNC\x00\x00\x00"


class TestOscServer(unittest.TestCase):
    def test_is_valid_request(self):
        self.assertTrue(
            osc_server._is_valid_request((b'#bundle\x00foobar',)))
        self.assertTrue(
            osc_server._is_valid_request((b'/address/1/2/3,foobar',)))
        self.assertFalse(
            osc_server._is_valid_request((b'',)))


class TestUDPHandler(unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.dispatcher = dispatcher.Dispatcher()
        # We do not want to create real UDP connections during unit tests.
        self.server = unittest.mock.Mock(spec=osc_server.BlockingOSCUDPServer)
        # Need to attach property mocks to types, not objects... weird.
        type(self.server).dispatcher = unittest.mock.PropertyMock(
            return_value=self.dispatcher)
        self.client_address = ("127.0.0.1", 8080)

    def test_no_match(self):
        mock_meth = unittest.mock.MagicMock()
        self.dispatcher.map("/foobar", mock_meth)
        osc_server._UDPHandler(
            [_SIMPLE_PARAM_INT_MSG, None], self.client_address, self.server)
        self.assertFalse(mock_meth.called)

    def test_match_with_args(self):
        mock_meth = unittest.mock.MagicMock()
        self.dispatcher.map("/SYNC", mock_meth, 1, 2, 3)
        osc_server._UDPHandler(
            [_SIMPLE_PARAM_INT_MSG, None], self.client_address, self.server)
        mock_meth.assert_called_with("/SYNC", [1, 2, 3], 4)

    def test_match_int9(self):
        mock_meth = unittest.mock.MagicMock()
        self.dispatcher.map("/debug", mock_meth)
        osc_server._UDPHandler(
            [_SIMPLE_PARAM_INT_9, None], self.client_address, self.server)
        self.assertTrue(mock_meth.called)
        mock_meth.assert_called_with("/debug", 9)

    def test_match_without_args(self):
        mock_meth = unittest.mock.MagicMock()
        self.dispatcher.map("/SYNC", mock_meth)
        osc_server._UDPHandler(
            [_SIMPLE_MSG_NO_PARAMS, None], self.client_address, self.server)
        mock_meth.assert_called_with("/SYNC")

    def test_match_default_handler(self):
        mock_meth = unittest.mock.MagicMock()
        self.dispatcher.set_default_handler(mock_meth)
        osc_server._UDPHandler(
            [_SIMPLE_MSG_NO_PARAMS, None], self.client_address, self.server)
        mock_meth.assert_called_with("/SYNC")


if __name__ == "__main__":
    unittest.main()



import unittest
import time

from pythonosc.parsing import ntp


class TestNTP(unittest.TestCase):
    """ TODO: Write real tests for this when I get time..."""

    def test_nto_to_system_time(self):
        unix_time = time.time()
        timestamp = ntp.system_time_to_ntp(unix_time)
        unix_time2 = ntp.ntp_to_system_time(timestamp)
        self.assertTrue(type(unix_time) is float)
        self.assertTrue(type(timestamp) is bytes)
        self.assertTrue(type(unix_time2) is float)
        self.assertAlmostEqual(unix_time, unix_time2, places=5)


if __name__ == "__main__":
    unittest.main()



import unittest

from pythonosc.dispatcher import Dispatcher, Handler


class TestDispatcher(unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.dispatcher = Dispatcher()

    def sortAndAssertSequenceEqual(self, expected, result):
        def sort(lst):
            return sorted(lst, key=lambda x: x.callback)

        return self.assertSequenceEqual(sort(expected), sort(result))

    def test_empty_by_default(self):
        self.sortAndAssertSequenceEqual([], self.dispatcher.handlers_for_address('/test'))

    def test_use_default_handler_when_set_and_no_match(self):
        handler = object()
        self.dispatcher.set_default_handler(handler)

        self.sortAndAssertSequenceEqual([Handler(handler, [])], self.dispatcher.handlers_for_address('/test'))

    def test_simple_map_and_match(self):
        handler = object()
        self.dispatcher.map('/test', handler, 1, 2, 3)
        self.dispatcher.map('/test2', handler)
        self.sortAndAssertSequenceEqual(
            [Handler(handler, [1, 2, 3])], self.dispatcher.handlers_for_address('/test'))
        self.sortAndAssertSequenceEqual(
            [Handler(handler, [])], self.dispatcher.handlers_for_address('/test2'))

    def test_example_from_spec(self):
        addresses = [
            "/first/this/one",
            "/second/1",
            "/second/2",
            "/third/a",
            "/third/b",
            "/third/c",
        ]
        for index, address in enumerate(addresses):
            self.dispatcher.map(address, index)

        for index, address in enumerate(addresses):
            self.sortAndAssertSequenceEqual(
                [Handler(index, [])], self.dispatcher.handlers_for_address(address))

        self.sortAndAssertSequenceEqual(
            [Handler(1, []), Handler(2, [])], self.dispatcher.handlers_for_address("/second/?"))

        self.sortAndAssertSequenceEqual(
            [Handler(3, []), Handler(4, []), Handler(5, [])],
            self.dispatcher.handlers_for_address("/third/*"))

    def test_do_not_match_over_slash(self):
        self.dispatcher.map('/foo/bar/1', 1)
        self.dispatcher.map('/foo/bar/2', 2)

        self.sortAndAssertSequenceEqual(
            [], self.dispatcher.handlers_for_address("/*"))

    def test_match_middle_star(self):
        self.dispatcher.map('/foo/bar/1', 1)
        self.dispatcher.map('/foo/bar/2', 2)

        self.sortAndAssertSequenceEqual(
            [Handler(2, [])], self.dispatcher.handlers_for_address("/foo/*/2"))

    def test_match_multiple_stars(self):
        self.dispatcher.map('/foo/bar/1', 1)
        self.dispatcher.map('/foo/bar/2', 2)

        self.sortAndAssertSequenceEqual(
            [Handler(1, []), Handler(2, [])], self.dispatcher.handlers_for_address("/*/*/*"))

    def test_match_address_contains_plus_as_character(self):
        self.dispatcher.map('/footest/bar+tender/1', 1)

        self.sortAndAssertSequenceEqual(
            [Handler(1, [])], self.dispatcher.handlers_for_address("/foo*/bar+*/*"))
        self.sortAndAssertSequenceEqual(
            [Handler(1, [])], self.dispatcher.handlers_for_address("/foo*/bar*/*"))

    def test_call_correct_dispatcher_on_star(self):
        self.dispatcher.map('/a+b', 1)
        self.dispatcher.map('/aaab', 2)
        self.sortAndAssertSequenceEqual(
            [Handler(2, [])], self.dispatcher.handlers_for_address('/aaab'))
        self.sortAndAssertSequenceEqual(
            [Handler(1, [])], self.dispatcher.handlers_for_address('/a+b'))

    def test_map_star(self):
        self.dispatcher.map('/starbase/*', 1)
        self.sortAndAssertSequenceEqual(
            [Handler(1, [])], self.dispatcher.handlers_for_address("/starbase/bar"))

    def test_map_root_star(self):
        self.dispatcher.map('/*', 1)
        self.sortAndAssertSequenceEqual(
            [Handler(1, [])], self.dispatcher.handlers_for_address("/anything/matches"))

    def test_map_double_stars(self):
        self.dispatcher.map('/foo/*/bar/*', 1)
        self.sortAndAssertSequenceEqual(
            [Handler(1, [])], self.dispatcher.handlers_for_address("/foo/wild/bar/wild"))
        self.sortAndAssertSequenceEqual(
            [], self.dispatcher.handlers_for_address("/foo/wild/nomatch/wild"))

    def test_multiple_handlers(self):
        self.dispatcher.map('/foo/bar', 1)
        self.dispatcher.map('/foo/bar', 2)
        self.sortAndAssertSequenceEqual(
            [Handler(1, []), Handler(2, [])], self.dispatcher.handlers_for_address("/foo/bar"))

    def test_multiple_handlers_with_wildcard_map(self):
        self.dispatcher.map('/foo/bar', 1)
        self.dispatcher.map('/*', 2)
        self.sortAndAssertSequenceEqual(
            [Handler(1, []), Handler(2, [])], self.dispatcher.handlers_for_address("/foo/bar"))

    def test_unmap(self):
        def dummyhandler():
            pass

        # Test with handler returned by map
        returnedhandler = self.dispatcher.map("/map/me", dummyhandler)
        self.sortAndAssertSequenceEqual([Handler(dummyhandler, [])], self.dispatcher.handlers_for_address("/map/me"))
        self.dispatcher.unmap("/map/me", returnedhandler)
        self.sortAndAssertSequenceEqual([], self.dispatcher.handlers_for_address("/map/me"))

        # Test with reconstructing handler
        self.dispatcher.map("/map/me/too", dummyhandler)
        self.sortAndAssertSequenceEqual([Handler(dummyhandler, [])],
                                        self.dispatcher.handlers_for_address("/map/me/too"))
        self.dispatcher.unmap("/map/me/too", dummyhandler)
        self.sortAndAssertSequenceEqual([], self.dispatcher.handlers_for_address("/map/me/too"))

    def test_unmap_exception(self):
        def dummyhandler():
            pass

        with self.assertRaises(ValueError) as context:
            self.dispatcher.unmap("/unmap/exception", dummyhandler)

        handlerobj = self.dispatcher.map("/unmap/somethingelse", dummyhandler())
        with self.assertRaises(ValueError) as context:
            self.dispatcher.unmap("/unmap/exception", handlerobj)


if __name__ == "__main__":
    unittest.main()



"""Maps OSC addresses to handler functions
"""

import collections
import logging
import re
import time
from pythonosc import osc_packet
from typing import overload, List, Union, Any, Generator, Tuple, Callable, Optional, DefaultDict
from pythonosc.osc_message import OscMessage


class Handler(object):
    """Wrapper for a callback function that will be called when an OSC message is sent to the right address.

    Represents a handler callback function that will be called whenever an OSC message is sent to the address this
    handler is mapped to. It passes the address, the fixed arguments (if any) as well as all osc arguments from the
    message if any were passed.
    """

    def __init__(self, _callback: Callable, _args: Union[Any, List[Any]],
                 _needs_reply_address: bool = False) -> None:
        """
        Args:
            _callback Function that is called when handler is invoked
            _args: Message causing invocation
            _needs_reply_address Whether the client's ip address shall be passed as an argument or not
       """
        self.callback = _callback
        self.args = _args
        self.needs_reply_address = _needs_reply_address

    # needed for test module
    def __eq__(self, other: Any) -> bool:
        return (type(self) == type(other) and
                self.callback == other.callback and
                self.args == other.args and
                self.needs_reply_address == other.needs_reply_address)

    def invoke(self, client_address: Tuple[str, int], message: OscMessage) -> None:
        """Invokes the associated callback function

        Args:
            client_address: Address match that causes the invocation
            message: Message causing invocation
       """
        if self.needs_reply_address:
            if self.args:
                self.callback(client_address, message.address, self.args, *message)
            else:
                self.callback(client_address, message.address, *message)
        else:
            if self.args:
                self.callback(message.address, self.args, *message)
            else:
                self.callback(message.address, *message)


class Dispatcher(object):
    """Maps Handlers to OSC addresses and dispatches messages to the handler on matched addresses

    Maps OSC addresses to handler functions and invokes the correct handler when a message comes in.
    """

    def __init__(self) -> None:
        self._map = collections.defaultdict(list)  # type: DefaultDict[str, List[Handler]]
        self._default_handler = None  # type: Optional[Handler]

    def map(self, address: str, handler: Callable, *args: Union[Any, List[Any]],
            needs_reply_address: bool = False) -> Handler:
        """Map an address to a handler

        The callback function must have one of the following signatures:

        ``def some_cb(address: str, *osc_args: List[Any]) -> None:``
        ``def some_cb(address: str, fixed_args: List[Any], *osc_args: List[Any]) -> None:``

        ``def some_cb(client_address: Tuple[str, int], address: str, *osc_args: List[Any]) -> None:``
        ``def some_cb(client_address: Tuple[str, int], address: str, fixed_args: List[Any], *osc_args: List[Any]) -> None:``

        Args:
            address: Address to be mapped
            handler: Callback function that will be called as the handler for the given address
            *args: Fixed arguements that will be passed to the callback function
            needs_reply_address: Whether the IP address from which the message originated from shall be passed as
                an argument to the handler callback

        Returns:
            The handler object that will be invoked should the given address match

        """
        # TODO: Check the spec:
        # http://opensoundcontrol.org/spec-1_0
        # regarding multiple mappings
        handlerobj = Handler(handler, list(args), needs_reply_address)
        self._map[address].append(handlerobj)
        return handlerobj

    @overload
    def unmap(self, address: str, handler: Handler) -> None:
        """Remove an already mapped handler from an address

        Args:
            address (str): Address to be unmapped
            handler (Handler): A Handler object as returned from map().
        """
        pass

    @overload
    def unmap(self, address: str, handler: Callable, *args: Union[Any, List[Any]],
              needs_reply_address: bool = False) -> None:
        """Remove an already mapped handler from an address

        Args:
            address: Address to be unmapped
            handler: A function that will be run when the address matches with
                the OscMessage passed as parameter.
            args: Any additional arguments that will be always passed to the
                handlers after the osc messages arguments if any.
            needs_reply_address: True if the handler function needs the
                originating client address passed (as the first argument).
        """
        pass

    def unmap(self, address, handler, *args, needs_reply_address=False):
        try:
            if isinstance(handler, Handler):
                self._map[address].remove(handler)
            else:
                self._map[address].remove(Handler(handler, list(args), needs_reply_address))
        except ValueError as e:
            if str(e) == "list.remove(x): x not in list":
                raise ValueError("Address '%s' doesn't have handler '%s' mapped to it" % (address, handler)) from e

    def handlers_for_address(self, address_pattern: str) -> Generator[Handler, None, None]:
        """Yields handlers matching an address


        Args:
            address_pattern: Address to match

        Returns:
            Generator yielding Handlers matching address_pattern
        """
        # First convert the address_pattern into a matchable regexp.
        # '?' in the OSC Address Pattern matches any single character.
        # Let's consider numbers and _ "characters" too here, it's not said
        # explicitly in the specification but it sounds good.
        escaped_address_pattern = re.escape(address_pattern)
        pattern = escaped_address_pattern.replace('\\?', '\\w?')
        # '*' in the OSC Address Pattern matches any sequence of zero or more
        # characters.
        pattern = pattern.replace('\\*', '[\w|\+]*')
        # The rest of the syntax in the specification is like the re module so
        # we're fine.
        pattern = pattern + '$'
        patterncompiled = re.compile(pattern)
        matched = False

        for addr, handlers in self._map.items():
            if (patterncompiled.match(addr)
                    or (('*' in addr) and re.match(addr.replace('*', '[^/]*?/*'), address_pattern))):
                yield from handlers
                matched = True

        if not matched and self._default_handler:
            logging.debug('No handler matched but default handler present, added it.')
            yield self._default_handler

    def call_handlers_for_packet(self, data: bytes, client_address: Tuple[str, int]) -> None:
        """Invoke handlers for all messages in OSC packet

        The incoming OSC Packet is decoded and the handlers for each included message is found and invoked.

        Args:
            data: Data of packet
            client_address: Address of client this packet originated from
        """

        # Get OSC messages from all bundles or standalone message.
        try:
            packet = osc_packet.OscPacket(data)
            for timed_msg in packet.messages:
                now = time.time()
                handlers = self.handlers_for_address(
                    timed_msg.message.address)
                if not handlers:
                    continue
                # If the message is to be handled later, then so be it.
                if timed_msg.time > now:
                    time.sleep(timed_msg.time - now)
                for handler in handlers:
                    handler.invoke(client_address, timed_msg.message)
        except osc_packet.ParseError:
            pass

    def set_default_handler(self, handler: Callable, needs_reply_address: bool = False) -> None:
        """Sets the default handler

        The default handler is invoked every time no other handler is mapped to an address.

        Args:
            handler: Callback function to handle unmapped requests
            needs_reply_address: Whether the callback shall be passed the client address
        """
        self._default_handler = None if (handler is None) else Handler(handler, [], needs_reply_address)



import socket
import unittest
from unittest import mock

from pythonosc import osc_message_builder
from pythonosc import udp_client


class TestUdpClient(unittest.TestCase):
    @mock.patch('socket.socket')
    def test_send(self, mock_socket_ctor):
        mock_socket = mock_socket_ctor.return_value
        client = udp_client.UDPClient('::1', 31337)

        msg = osc_message_builder.OscMessageBuilder('/').build()
        client.send(msg)

        self.assertTrue(mock_socket.sendto.called)
        mock_socket.sendto.assert_called_once_with(msg.dgram, ('::1', 31337))


class TestSimpleUdpClient(unittest.TestCase):
    def setUp(self):
        self.patcher = mock.patch('pythonosc.udp_client.OscMessageBuilder')
        self.patcher.start()
        self.builder = udp_client.OscMessageBuilder.return_value
        self.msg = self.builder.build.return_value
        self.client = mock.Mock()

    def tearDown(self):
        self.patcher.stop()

    def test_send_message_calls_send_with_msg(self):
        udp_client.SimpleUDPClient.send_message(self.client, '/address', 1)
        self.client.send.assert_called_once_with(self.msg)

    def test_send_message_calls_add_arg_with_value(self):
        udp_client.SimpleUDPClient.send_message(self.client, '/address', 1)
        self.builder.add_arg.assert_called_once_with(1)

    def test_send_message_calls_add_arg_once_with_string(self):
        udp_client.SimpleUDPClient.send_message(self.client, '/address', 'hello')
        self.builder.add_arg.assert_called_once_with('hello')

    def test_send_message_calls_add_arg_multiple_times_with_list(self):
        udp_client.SimpleUDPClient.send_message(self.client, '/address',
                                                [1, 'john', True])
        self.assertEqual(self.builder.add_arg.call_count, 3)


if __name__ == "__main__":
    unittest.main()



"""Use OSC packets to parse incoming UDP packets into messages or bundles.

It lets you access easily to OscMessage and OscBundle instances in the packet.
"""

import time

from pythonosc.parsing import osc_types
from pythonosc import osc_bundle
from pythonosc import osc_message

from typing import Union, List, NamedTuple

# A namedtuple as returned my the _timed_msg_of_bundle function.
# 1) the system time at which the message should be executed
#    in seconds since the epoch.
# 2) the actual message.
TimedMessage = NamedTuple('TimedMessage', [
    ('time', float),
    ('message', osc_message.OscMessage),
])


def _timed_msg_of_bundle(bundle: osc_bundle.OscBundle, now: float) -> List[TimedMessage]:
    """Returns messages contained in nested bundles as a list of TimedMessage."""
    msgs = []
    for content in bundle:
        if type(content) is osc_message.OscMessage:
            if (bundle.timestamp == osc_types.IMMEDIATELY or bundle.timestamp < now):
                msgs.append(TimedMessage(now, content))
            else:
                msgs.append(TimedMessage(bundle.timestamp, content))
        else:
            msgs.extend(_timed_msg_of_bundle(content, now))
    return msgs


class ParseError(Exception):
    """Base error thrown when a packet could not be parsed."""


class OscPacket(object):
    """Unit of transmission of the OSC protocol.

    Any application that sends OSC Packets is an OSC Client.
    Any application that receives OSC Packets is an OSC Server.
    """

    def __init__(self, dgram: bytes) -> None:
        """Initialize an OdpPacket with the given UDP datagram.

        Args:
          - dgram: the raw UDP datagram holding the OSC packet.

        Raises:
          - ParseError if the datagram could not be parsed.
        """
        now = time.time()
        try:
            if osc_bundle.OscBundle.dgram_is_bundle(dgram):
                self._messages = sorted(
                    _timed_msg_of_bundle(osc_bundle.OscBundle(dgram), now),
                    key=lambda x: x.time)
            elif osc_message.OscMessage.dgram_is_message(dgram):
                self._messages = [TimedMessage(now, osc_message.OscMessage(dgram))]
            else:
                # Empty packet, should not happen as per the spec but heh, UDP...
                raise ParseError(
                    'OSC Packet should at least contain an OscMessage or an '
                    'OscBundle.')
        except (osc_bundle.ParseError, osc_message.ParseError) as pe:
            raise ParseError('Could not parse packet %s' % pe)

    @property
    def messages(self) -> List[TimedMessage]:
        """Returns asc-time-sorted TimedMessages of the messages in this packet."""
        return self._messages
